/*
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <oleauto.h>
#include "initguid.h"
#include "pimstore.h"
#include <stdio.h>

#include "base/startcmd.h"
#include "pim/maincpp.h"
#include "pim/Utils.h"
#include "pim/SettingFunctions.h"
#include "client/Sync4jClient.h"
#include "pim/ContactSyncSource.h"
#include "pim/CalendarSyncSource.h"
#include "pim/NoteSyncSource.h"
#include "pim/TaskSyncSource.h"
#include "pim/BriefcaseSyncSource.h"

#include "pim/mail/MailSyncSource.h"
#include "client/MailSourceManagementNode.h"
#include "filter/SourceFilter.h"
#include "filter/ClauseUtil.h"

#include "spdm/DMTree.h"
#include "spdm/DMTreeFactory.h"
#include "client/SourceManagementNode.h"
#include "syncml/formatter/Formatter.h"

#include "Notify.h"
#include "notify/timed_msgbox.h"

#include "pim/difftime.h"

/*      
 * Entry point from VB application. 
 * Accept and manage Variant array of String, long and int. 
 * The entry array is a String array in which elements rapresent SyncSource name.
 * Now accepted and managed element names are "contacts" and "appointments"
 * Array contains at least one element
 * @param vbErrorMsg   : the variable to store the message to return to vb GUI interface
 * @param path       : the install path of the SyncClient PIM application on the device
 */

DWORD WINAPI synchronization (wchar_t* vbErrorMsg, wchar_t* path) {
    
    wchar_t CONTACTS    [15];
    wchar_t APPOINTMENTS[15];       
    wchar_t TASKS       [15];
    wchar_t FAVORITES   [15];
    wchar_t FILES       [15];
    wchar_t NOTES       [15];
    wchar_t MAILS       [15];    
    wchar_t errorMessage[1024];
    int rc          = 0;
    lastErrorCode   = 0;
    ArrayList syncSources;
    char* value;
    errorMessage[0] = 0;
    SyncSource** ssArray = NULL;
    int ret = 0;
    BOOL isMailToSync = FALSE;
    int mailSourceIndex = -1; 

    wsprintf(CONTACTS,      CONTACTS_NAME    );
    wsprintf(APPOINTMENTS,  APPOINTMENTS_NAME);
    wsprintf(TASKS,         TASKS_NAME       );
    wsprintf(FAVORITES,     FAVORITES_NAME   );
    wsprintf(FILES,         FILES_NAME       );
    wsprintf(NOTES,         NOTES_NAME       );
    wsprintf(MAILS,         MAILS_NAME       );

    //LOG.setLogPath(path);
    
    // to remember code only...
    // int constFolderCalendar  =  9; 
    // int constFolderContacts  = 10;
    // int constFolderTasks     = 13;

    //
    // custom values
    //
    // int constFiles       = 600;
    // int constNotes       = 601;
    // int constFavorites   = 602; 
           
    // 
    // The list of the sync sources to sync
    //
    DMTree* dmt = DMTreeFactory::getDMTree("");		
	ManagementNode* node = NULL;
    
    // contact
    node = dmt->getManagementNode(CONTACTS_CONTEXT);
	value = node->getPropertyValue("sync");
    if (strcmp(value, "none") != 0) {
        ContactSyncSource contactSource = ContactSyncSource(CONTACTS, OL_CONTACTS);     
        
        contactSource.setPath(path);
        syncSources.add(contactSource);
    }
    if (node) { delete node ; node = NULL; }

    // calendar
    node = dmt->getManagementNode(APPOINTMENTS_CONTEXT);	
	value = node->getPropertyValue("sync");
    if (strcmp(value, "none") != 0) {
        CalendarSyncSource  s  = CalendarSyncSource (APPOINTMENTS,     OL_CALENDAR);         
        s.setPath(path);
        syncSources.add(s);
    }
    if (node) { delete node ; node = NULL; }

    // task
    node = dmt->getManagementNode(TASKS_CONTEXT);
	value = node->getPropertyValue("sync");
    if (strcmp(value, "none") != 0) {
        TaskSyncSource  s  = TaskSyncSource (TASKS,     OL_TASK);         
        s.setPath(path);
        syncSources.add(s);
    }
    if (node) { delete node ; node = NULL; }

     // briefcase
    node = dmt->getManagementNode(BRIEFCASE_CONTEXT);
	value = node->getPropertyValue("sync");
    if (strcmp(value, "none") != 0) {
        BriefcaseSyncSource  s  = BriefcaseSyncSource (FILES,     OL_BRIEFCASE);         
        s.setPath(path);
        if (node) { delete node ; node = NULL; }

        node = dmt->getManagementNode(BRIEFCASE_CONTEXT);
        value = node->getPropertyValue("dir");
        wchar_t *dir = toWideChar(value);
        s.setDir(dir);
        syncSources.add(s);
        delete [] dir;
    }
    if (node) { delete node ; node = NULL; }

     // note
    node = dmt->getManagementNode(NOTES_CONTEXT);
	value = node->getPropertyValue("sync");
    if (strcmp(value, "none") != 0) {
        NoteSyncSource  s  = NoteSyncSource (NOTES,     OL_NOTES);         
        s.setPath(path);
        if (node) { delete node ; node = NULL; }
        node = dmt->getManagementNode(NOTES_CONTEXT);
        wchar_t *dir = toWideChar(node->getPropertyValue("dir"));
        s.setDir(dir);
        syncSources.add(s);
    }
    if (node) { delete node ; node = NULL; }


    // mail
    MailSourceManagementNode m = MailSourceManagementNode(SOURCES_CONTEXT, "mails");   
    MailSyncSourceConfig sc;    
    wchar_t t[12];
    sc = m.getMailSourceConfig(true);    
    MailSyncSource mailSource = MailSyncSource (MAILS); 
    
    if (strcmp(sc.getSync(), "none") != 0) {
        getFolderToSync(sc, t);    
    
        int days = sc.getDownloadAge();
        wchar_t* downloadAge = getDayBefore(days);

        SourceFilter* filter = 
            ClauseUtil::createSourceFilter(downloadAge, sc.getBodySize(),sc.getAttachSize());        

        mailSource.setPath(path);
        mailSource.setFolderToSync(t);
        mailSource.setFilter(filter);

        syncSources.add(mailSource);
        isMailToSync = TRUE;
        mailSourceIndex = syncSources.size()-1;        

    }   

    if (syncSources.size() == 0) {
        wsprintf(errorMessage, TEXT("No sources to sync"));
    
    } else {
                
        Sync4jClient& s4j = Sync4jClient::getInstance();
        s4j.setDMConfig(APPLICATION_URI);   
        LogLevel ll = getLOGLevel();
        LOG.setLevel(ll);        

        //ret = s4j.sync(syncSources);         
        
        int l = syncSources.size();
        const int ml = 10;        
        SyncSource* array[ml];    
        for (int i=0; (i<l) && (i<ml-1); ++i) {
            array[i] = (SyncSource*)syncSources[i];
        }
        array[l] = NULL;
        ret = s4j.sync(array);
        
        if (ret == 0) {
        // If the mail has been synchronized
            if (mailSourceIndex >= 0) {
                MailSyncSource *s = (MailSyncSource *)array[mailSourceIndex];

                // Re-sync if a message was in outbox, to get the result from the server
                if ( s->getIsMailInOutbox() ) {
                    // 1) If the mail source is in a position different from 0, copy
                    //    it to the first position
                    if(mailSourceIndex > 0)
                        array[0] = s;
                    // 2) Truncate the array
                    array[1] = NULL;
                    // 3) Do the second sync for mail source
                    s4j.setDMConfig(APPLICATION_URI);
                    ret = s4j.sync(array);
                }
                // Alert user if there is new mail
                if (s->getIsMailInInbox()) {
                    MessageBeep(MB_ICONASTERISK);
                }
                if (s->getFailedSendMailInOutbox()) {
                    TimedMessageBox(NULL, OUT_MESSAGE_NOT_SENT, TEXT("Alert"),
                        MB_OK | MB_ICONINFORMATION, OUT_MESSAGE_TIMEOUT*1000);
                }
            }
        }
        //
        // If there is a mail in Outbox folder, the sync is performed twice.
        // 
       
        Sync4jClient::dispose();
    }
    
    //
    // NOTE: when passing a string from VB to a DLL, it looks like
    // VB does not want that the last chararcter in the string is
    // the terminator. Therefore we change it to a space. However,
    // IT IS VERY IMPORTANT that the string allocated by the caller
    // less than the length of errorMessage. 
    //
    wsprintf(vbErrorMsg, errorMessage);
    vbErrorMsg[wcslen(vbErrorMsg)] = L' ';
    vbErrorMsg[wcslen(vbErrorMsg)+1] = 0;    
    logInfo(TEXT("Sync ended."));
    
    return ret;

}

/*      
 * Entry point to kickoff the synchronization process, specifying which
 * sources to synchronize.
 *
 * @param path    : the install path of the SyncClient PIM application on 
 *                  the device
 * @param sources : NULL terminated array of strings containing the source
 *                  name of the sources to synchronize
 */

DWORD WINAPI synchronizeMultipleSources (const wchar_t* path, const wchar_t** sources) {
    
    if (sources == 0) {
        return 0;
    }

    int ret = 0;
    DMTree* dmt = DMTreeFactory::getDMTree("");		
	ManagementNode* node = NULL;
    int i=0;
    ArrayList syncSources;
    int mailSourceIndex = -1;
    
    //LOG.setLogPath(path);

    for (i=0; sources[i]; i++) {
        if (wcscmp(sources[i], CONTACTS_NAME) == 0) {
            ContactSyncSource s(CONTACTS_NAME, OL_CONTACTS);
            s.setPath(path);
            syncSources.add(s);
        }
        else if (wcscmp(sources[i], APPOINTMENTS_NAME) == 0) {
            CalendarSyncSource s(APPOINTMENTS_NAME, OL_CALENDAR);
            s.setPath(path);
            syncSources.add(s);
        }
        else if (wcscmp(sources[i], TASKS_NAME)        == 0) {
            TaskSyncSource  s(TASKS_NAME, OL_TASK);         
            s.setPath(path);
            syncSources.add(s);
        }
        else if (wcscmp(sources[i], FAVORITES_NAME)    == 0) {
/*
            FavoritesSyncSource s(FAVORITES_NAME, OL_FAVORITES);
            s.setPath(path);
            syncSources.add(s);
*/
        }
        else if (wcscmp(sources[i], FILES_NAME)        == 0) {   
            BriefcaseSyncSource  s(FILES_NAME, OL_BRIEFCASE);         
            s.setPath((wchar_t*)path);
            
            node = dmt->getManagementNode(BRIEFCASE_CONTEXT);
            if (node) {
                wchar_t* value = toWideChar(node->getPropertyValue("dir"));
                s.setDir(value);
                syncSources.add(s);
                delete node;
                delete [] value;
            }
            else {
                logError(TEXT("Can't access management node %s"), BRIEFCASE_CONTEXT);
            }
        } 
        else if (wcscmp(sources[i], NOTES_NAME) == 0) {
            NoteSyncSource s(NOTES_NAME, OL_NOTES);         
            s.setPath(path);
            
            node = dmt->getManagementNode(NOTES_CONTEXT);
            wchar_t* value = toWideChar(node->getPropertyValue("dir"));
            s.setDir(value);
            syncSources.add(s);
            if (node) {
                delete node ; node = NULL;
            }
        }
        else if (wcscmp(sources[i], MAILS_NAME) == 0) {
            MailSourceManagementNode m(SOURCES_CONTEXT, "mails");   
            MailSyncSourceConfig sc;    
            wchar_t t[12];
            sc = m.getMailSourceConfig(true);    
       
            getFolderToSync(sc, t);    
    
            int days = sc.getDownloadAge();
            wchar_t* downloadAge = getDayBefore(days);

            SourceFilter* filter = 
                ClauseUtil::createSourceFilter(downloadAge, sc.getBodySize(),sc.getAttachSize());

            MailSyncSource mailSource(MAILS_NAME);
            
            mailSource.setPath(path);
            mailSource.setFolderToSync(t);
            mailSource.setFilter(filter);

            syncSources.add(mailSource);

            mailSourceIndex = syncSources.size()-1;
        }
    }

    //
    // Now we can create the Sync4jClient object and kick off the sync
    //
    Sync4jClient& s4j = Sync4jClient::getInstance();
    s4j.setDMConfig(APPLICATION_URI);
    LogLevel ll = getLOGLevel();
    LOG.setLevel(ll);

    //ret = s4j.sync(syncSources); 
    int l = syncSources.size();
    const int ml = 10;        
    SyncSource* s_array[ml];
    for (i=0; (i<l) && (i<ml-1); ++i) {
        s_array[i] = (SyncSource*)syncSources[i];
    }
    s_array[i] = NULL;
    ret = s4j.sync(s_array);

    if (ret == 0) {
        // If the mail has been synchronized
        if (mailSourceIndex >= 0) {
            MailSyncSource *s = (MailSyncSource *)s_array[mailSourceIndex];

            // Re-sync if a message was in outbox, to get the result from the server
            if ( s->getIsMailInOutbox() ) {
                // 1) If the mail source is in a position different from 0, copy
                //    it to the first position
                if(mailSourceIndex > 0)
                    s_array[0] = s;
                // 2) Truncate the array
                s_array[1] = NULL;
                // 3) Do the second sync for mail source
                s4j.setDMConfig(APPLICATION_URI);
                ret = s4j.sync(s_array);
            }
            // Alert user if there is new mail
            if (s->getIsMailInInbox()) {
                MessageBeep(MB_ICONASTERISK);
            }
            if (s->getFailedSendMailInOutbox()) {
                TimedMessageBox(NULL, OUT_MESSAGE_NOT_SENT, TEXT("Alert"),
                    MB_OK | MB_ICONINFORMATION, OUT_MESSAGE_TIMEOUT*1000);
            }
        }
    }
    Sync4jClient::dispose();
    logDebug(TEXT("Sync ended multiple sources."));
    return ret;
}

/**
 * Return TRUE if the process is started
 */
BOOL startPolling()
{    
    return (startcmd(SYNCAPP, APP_RUN_AT_TIME) > 0);
}

/*
 * Return 0: process terminated successfully
 *        >0: error code from child process    
 *       -1: no such process/invalid handle
 *       -2: timeout
 *       -3: can't get child exit code
 *       -5: can't run the app
 *       -6: the application is stopped by user (used into startSyncPid and checkPid)
 */

int startSync(time_t timeout)
{
    if (!timeout)
        timeout=5*60; // default timeout 5 minutes

    unsigned long pid = startcmd(SYNCAPP, TEXT("once"));
   
    if (pid) {
        return waitProcess(pid, timeout*1000);        
    }
    else {
        return -5;        
    }
    
}

/*
* Used from evb interface. It returns:
* . 1000 if the handle is NULL; so the processId is not valid or not assiciated to this process
* . STILL_ACTIVE (259) the process is still active.
* . a different value indicates the proper exit value.
*/

int checkPid(long pid) {

    unsigned long use = pid;
    DWORD exitCode = 1000;
    wchar_t t[128];
    
    wsprintf(t, TEXT("%lu"), use);    
    
    HANDLE phandle = OpenProcess(0, FALSE, use);  
    
    if (phandle) {
        BOOL f = GetExitCodeProcess(phandle, &exitCode);
        CloseHandle(phandle);        
    } 
    
    return exitCode;
    
}

unsigned long startSyncPid(time_t timeout)
{
    if (!timeout)
        timeout=5*60; // default timeout 5 minutes

    unsigned long pid = startcmd(SYNCAPP, TEXT("once"));
    
    if (pid) {
        return pid;
    }
    else {
        return -5;
    }
}


/*      
 * Entry point to kickoff the synchronization process, both specifying the 
 * sources or not. The default is to make all the configured sources.
 *
 * @param path    : the install path of the SyncClient PIM application on 
 *                  the device
 * @param sources : NULL terminated array of strings containing the source
 *                  name of the sources to synchronize
 */

DWORD WINAPI synchronize (wchar_t *errmsg, const wchar_t* path, const wchar_t** sources)
{
    wcscpy(errmsg, TEXT("Ok."));
    LOG.reset();
    if (sources)
        return synchronizeMultipleSources(path, sources);
    else
        return synchronization(errmsg, (wchar_t *)path);
}

/*
    LOG_LEVEL_NONE  = 0
    LOG_LEVEL_INFO  = 1
    LOG_LEVEL_DEBUG = 2
*/

LogLevel getLOGLevel() {
    
    wchar_t level[64];
    memset(level, 0, 64*sizeof(wchar_t));
    LogLevel ret = LOG_LEVEL_NONE;
    long l = 0;

    getClientConfigurationInternal(NULL, PROPERTY_SPDM_LOG_LEVEL, level, FALSE);
    
    if (wcslen(level) != 0)
        l = wcstol(level, NULL, 0);
    
    ret = (LogLevel)l;

    return ret;    
}

/*
void testInsertMail(StringBuffer s) {
    MailServer2Client* mailServer2Client = new MailServer2Client();
    MailClientData m;
    m.setParent(TEXT("I"));
    m.emailData->parse(s.c_str());
    
    mailServer2Client->setFolderToSync(TEXT("I"));
    mailServer2Client->addMessage(&m);

}
*/