/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pim/mail/MailServer2Client.h"
#include "base/util/utils.h"
#include "base/util/BasicTime.h"
#include "pim/difftime.h"

#include "base/Log.h"

//------------------------------------------------------------- Typedef

enum {
	eeePR_ENTRYID,
	eeePR_MESSAGE_FLAGS,
    eeePR_LAST_MODIFICATION_TIME,
	eeeNUM_COLS};

// These tags represent the message information we would like to pick up
static SizedSPropTagArray(eeeNUM_COLS, endSync) = 
{	
	eeeNUM_COLS,
    PR_ENTRYID,                       // entry id
	PR_MESSAGE_FLAGS,    
    PR_LAST_MODIFICATION_TIME        //

};

 enum {
	ePR_ENTRYID,
	ePR_MESSAGE_FLAGS,
    //MM        
    ePR_LAST_MODIFICATION_TIME,
    ePR_SUBJECT,
    ePR_SUBJECT_PREFIX,
    ePR_MESSAGE_SIZE,    
    ePR_SENDER_NAME,        
    ePR_SENDER_EMAIL_ADDRESS,    
    ePR_HASATTACH,
    //MM
	NUM_COLS};

// These tags represent the message information we would like to pick up
static SizedSPropTagArray(NUM_COLS, properties) = 
{	
	NUM_COLS,
    PR_ENTRYID,                         // entry id
	PR_MESSAGE_FLAGS,
    //MM        
    PR_LAST_MODIFICATION_TIME,          //
    PR_SUBJECT,                         // mail Subject
    PR_SUBJECT_PREFIX,
    PR_MESSAGE_SIZE,                    // sum of the size of every property of the object
    PR_SENDER_NAME,                     // "Marco "
    PR_SENDER_EMAIL_ADDRESS,            // From  "Marco <magi@libero.it>    
    PR_HASATTACH

};

enum {
	eePR_ENTRYID,
    eePR_LAST_MODIFICATION_TIME,
	eePR_DISPLAY_NAME,     
    eePR_MESSAGE_FLAGS,
    eeNUM_COLS};

// These tags represent the message information we would like to pick up
static SizedSPropTagArray(eeNUM_COLS, Columns) = 
{	
	eeNUM_COLS,
	PR_ENTRYID,
    PR_LAST_MODIFICATION_TIME,
	PR_DISPLAY_NAME,
    PR_MESSAGE_FLAGS
};

//----------------------------------------------------- Local functions
static ULONG CountRecipsInString(LPCWSTR psz)
{
  ULONG cRecips = 0;

  if (psz != NULL)
  {
    if (*psz != '\0')
      ++cRecips;

    while (*psz != L'\0')
    {
      if (*psz == L';' || *psz == L',') 
        ++cRecips;
      ++psz;
    }
  }

  return cRecips;
}


/*
* Check if the t char starts with I/, O/, D/, T/, S/
*/
static inline bool startWithFolder(const char* t) {
    bool ret = false;
    if (strstr(t, "I/") != NULL ||
        strstr(t, "O/") != NULL ||
        strstr(t, "D/") != NULL ||
        strstr(t, "S/") != NULL ||
        strstr(t, "T/") != NULL )
        
        ret = true;
    
    return ret;
}

//------------------------------------------------------ Constructors

MailServer2Client::MailServer2Client()
{
    server_folderToSync = NULL;
}

MailServer2Client::~MailServer2Client()
{
    if (server_folderToSync) {
        delete [] server_folderToSync; server_folderToSync = NULL;
    }
}

//----------------------------------------------------------- Methods

HRESULT MailServer2Client::AddAttachment(LPMESSAGE pmsg, MailClientData* m)
{
    LOG.debug("Enter into AddAttachment function");
	HRESULT hr = S_OK;
    BodyPart* bodyPart = NULL;
    bodyPart = m->getEmailData()->getEmailItem().getFirstAttachment();

    if (bodyPart == NULL) {
        LOG.debug("Exiting from AddAttachment function: no body part found");
        return hr;
    }

    do {
        
        wchar_t* tmpFileName = toWideChar(bodyPart->getContent());
        wchar_t* filename    = toWideChar(bodyPart->getFilename());

        SPropValue  prop[3];
		IAttach*    pAttachment = NULL;
		ULONG       ulAttachmentNum = 0;
		FILE*       fileAttach = NULL;;
        
        memset(prop, 0, sizeof(prop));
        fileAttach = _wfopen(tmpFileName, TEXT("rb"));	
        
        if (fileAttach == NULL) {
            hr = -1;
            break;
        }
        
        hr = pmsg->CreateAttach(NULL, 0, &ulAttachmentNum, &pAttachment);
		EXIT_ON_FAILED(hr);

		// Set the initial properties on the attachment 
        // no data actually there yet
		prop[0].ulPropTag = PR_ATTACH_METHOD;
		prop[0].Value.ul = ATTACH_BY_VALUE;

		prop[1].ulPropTag = PR_ATTACH_DATA_BIN;
        
        prop[2].ulPropTag = PR_ATTACH_FILENAME;
		prop[2].Value.LPSZ = filename;
        
        hr = pAttachment->SetProps(3, prop, NULL);
		EXIT_ON_FAILED(hr);

		// Now get the data property and fill it in
		//
		IStream* pStream;

		hr = pAttachment->OpenProperty (PR_ATTACH_DATA_BIN,
									 NULL,
									 0,
									 MAPI_MODIFY | MAPI_CREATE,
									 (LPUNKNOWN *)&pStream);
		EXIT_ON_FAILED(hr);
        
        //DWORD cbFile = GetFileSize(fileAttach, NULL);
        DWORD cbFile = fgetsize(fileAttach);
        if (cbFile == 0xFFFFFFFF) // error
            break;
        
		ULONG        cbWritten;
		ULONG        cbReadSize;
		const ULONG  ChunkSize = 4096;
		BYTE         Buffer[ChunkSize];

		do {
			// cbReadSize = fileAttach.Read(Buffer, ChunkSize);
            cbReadSize = fread(Buffer, sizeof(BYTE), ChunkSize, fileAttach ); 
                
			if (cbReadSize > 0)
			{
				hr = pStream->Write(Buffer, cbReadSize, &cbWritten);
			}
		} while (!FAILED(hr) && (cbReadSize == ChunkSize));

		fflush(fileAttach);
        fclose(fileAttach);
		pStream->Commit(STGC_DEFAULT);
		pStream->Release();

		// Set the size property
		//
		prop[0].ulPropTag = PR_ATTACH_SIZE;
		prop[0].Value.ul = cbFile + sizeof(prop[2]);
		hr = pAttachment->SetProps(1, prop, NULL);                
        
        if (tmpFileName != NULL) {
            DeleteFile(tmpFileName);
        }


    } while ((bodyPart = m->getEmailData()->getEmailItem().getNextAttachment()) != NULL);

FuncExit:  
    LOG.debug("Exiting from AddAttachment function");
    return hr;   
}


static TCHAR szSMTP[] = L"SMTP";

void AddRecipients(LPWSTR pszSrc, ULONG ulType, ADRLIST* plst, LPWSTR& pszLocalRecips, LPSPropValue& pval)
{   
    LOG.debug("Enter into AddRecipients function");
    if (pszSrc == NULL) {
        LOG.debug("Exiting from AddRecipients function: no recipients");
        return;
    }
    LPWSTR psz = pszSrc;
    while (*psz != L'\0') {
        while (*psz == L' ' && *psz != '\0')
            ++psz;

        if (*psz == L'\0')
            break;

        LPWSTR pszEnd = psz;
        // while (*pszEnd != L' ' && (*pszEnd != ';' || *pszEnd != ',') && *pszEnd != '\0')
        while (*pszEnd != ',' && *pszEnd != '\0')
            ++pszEnd;

        int cch = pszEnd - psz;

        wcsncpy(pszLocalRecips, psz, cch);
        *(pszLocalRecips + cch) = L'\0';

        plst->aEntries[plst->cEntries].cValues = 3;
        plst->aEntries[plst->cEntries].rgPropVals = pval;

        // Set the type (To, Cc, Bcc)...
        pval->ulPropTag = PR_RECIPIENT_TYPE;
        pval->Value.ul = ulType;
        ++pval;

        // Set the address type (we only do SMTP)...
        pval->ulPropTag = PR_ADDRTYPE;
        pval->Value.lpszW = szSMTP;
        ++pval;

        // Set the address...
        pval->ulPropTag = PR_EMAIL_ADDRESS;
        pval->Value.lpszW = pszLocalRecips;
        ++pval;

        ++plst->cEntries;

        pszLocalRecips += wcslen(pszLocalRecips) + 1;

        if (*pszEnd != L'\0')
            ++pszEnd;

        psz = pszEnd;
    }
    LOG.debug("Exiting from AddRecipients function");
}

HRESULT MailServer2Client::SetMessageRecipients (LPMESSAGE pmsg, MailClientData* m) {

    LOG.debug("Enter into SetMessageRecipients function");
    ULONG cRecipients = 0;
    ULONG cchTotalRecips = 0;

    ADRLIST* plst = NULL;
    LPSPropValue pval = NULL;
    BYTE* pb = NULL;
    LPWSTR pszLocalRecips;
    HRESULT hr;
    wchar_t *to = NULL, *cc = NULL, *bcc = NULL; 
    if (m->getEmailData()->getEmailItem().getTo()) {
        to = toWideChar(m->getEmailData()->getEmailItem().getTo());
        cRecipients += CountRecipsInString(to);
        cchTotalRecips += wcslen(to) + 3;
    }

    if (m->getEmailData()->getEmailItem().getCc()) {
        cc = toWideChar(m->getEmailData()->getEmailItem().getCc());
        cRecipients += CountRecipsInString(cc);
        cchTotalRecips += wcslen(cc) + 3;
    }

    if (m->getEmailData()->getEmailItem().getBcc()) {
        bcc = toWideChar(m->getEmailData()->getEmailItem().getBcc());
        cRecipients += CountRecipsInString(bcc);
        cchTotalRecips += wcslen(bcc) + 3;
    }
  
    //
    // Allocate one big block of memory to hold all the strings.
    // The block is arranged as follows:
    //
    // ADRLIST        SPropValue's     Copy of addresses
    // |---------------------|--------------------|-------------------|
    //

    DWORD cb = sizeof(ADRLIST) + ((sizeof(ADRENTRY) + sizeof(SPropValue) * 3) * cRecipients);

    pb = new BYTE[cb + (cchTotalRecips * sizeof(WCHAR))];
    if (pb == NULL)
    return E_OUTOFMEMORY;

    ZeroMemory(pb, cb + (cchTotalRecips * sizeof(WCHAR)));

    plst = (ADRLIST*) pb;

    pszLocalRecips = (LPWSTR) (pb + cb);

    pb += sizeof(ADRLIST) + (sizeof(ADRENTRY) * cRecipients);

    pval = (SPropValue*) pb;

    AddRecipients(to, MAPI_TO, plst, pszLocalRecips, pval);
    AddRecipients(cc, MAPI_CC, plst, pszLocalRecips, pval);
    AddRecipients(bcc, MAPI_BCC, plst, pszLocalRecips, pval);

    hr = pmsg->ModifyRecipients(MODRECIP_ADD, plst);

    delete[] (BYTE*) plst;
    if (to) {delete [] to; to = NULL;}
    if (cc) {delete [] cc; cc = NULL;}
    if (bcc) {delete [] bcc; bcc = NULL;}

    LOG.debug("Exiting from SetMessageRecipients function");
    return hr;
}

void MailServer2Client::bt2ft(const BasicTime& b, FILETIME& f)
{
    SYSTEMTIME local, s1;

    local.wYear         = b.getYear();
    local.wMonth        = b.getMonth();
    local.wDay          = b.getDay();
    local.wDayOfWeek    = DayOfWeek(local.wYear, local.wMonth, local.wDay);    // b.getWeekday(); // 0 sunday, 1monday
    local.wHour         = b.getHour();
    local.wMinute       = b.getMin();
    local.wSecond       = b.getSec();
    local.wMilliseconds = 0;

    s1 = DT_AddDiff(nano100SecInHour, -(b.getTzHour()), &local); 
    SystemTimeToFileTime(&s1, &f); 

}

HRESULT MailServer2Client::SetMessageProps(LPMESSAGE pmsg, MailClientData* m) {
  
    LOG.debug("Enter into SetMessageProps function");
    SPropValue rgprops[8] = {0};
    ULONG    cProps   = 0;
    HRESULT   hr;
    BodyPart bp;
    wchar_t *subj = NULL, *from = NULL, 
            *messId = NULL, *headers = NULL;    

    LPSTREAM  pstm    = NULL;
    ULONG messageFlag = (ULONG) 0x00000000;        

    char* h = NULL;
    // Set the recipients up.
    hr = SetMessageRecipients(pmsg, m);
    EXIT_ON_FAILED(hr);
    
    MailMessage &msg = m->getEmailData()->getEmailItem();
    
    if (msg.attachmentCount() > 0) {
        rgprops[cProps].ulPropTag = PR_HASATTACH;
        rgprops[cProps].Value.ul = msg.attachmentCount();
        ++cProps;
        //messageFlag = messageFlag | MSGFLAG_HASATTACH;
    }

    //
    // Set this flag only for Inbox and Draft. This shows the From in the
    // visualization of the mail into the pocket pc    
    
    if (wcscmp(m->getParent(), TEXT("I")) == 0 || 
        wcscmp(m->getParent(), TEXT("T")) == 0) {
        
        messageFlag = messageFlag | MSGFLAG_UNMODIFIED ;

    } else if (wcscmp(m->getParent(), TEXT("D")) == 0) {

        messageFlag = messageFlag | MSGFLAG_FROMME | MSGFLAG_UNSENT;

    } else if (wcscmp(m->getParent(), TEXT("S")) == 0) {

        messageFlag = messageFlag | MSGFLAG_FROMME ;
    }
     
    //
    // To be an editable message the MSGFLAG_FROMME has to be set.
    // surely the mail into draft folder has to be flagged MSGFLAG_FROMME.
    // X-Funambol-isDraft
            
    if (m->getEmailData()->getRead())
        messageFlag = messageFlag | MSGFLAG_READ;
  
    
    // Set the flags and a subject if they exist.
    rgprops[cProps].ulPropTag = PR_MESSAGE_FLAGS;
    rgprops[cProps].Value.ul = messageFlag;  
    // rgprops[cProps].Value.ul = MSGFLAG_FROMME | MSGFLAG_UNSENT;  
    // rgprops[cProps].Value.ul = MSGFLAG_READ;
    ++cProps;
  
    rgprops[cProps].ulPropTag = PR_MSG_STATUS;
    rgprops[cProps].Value.ul = MSGSTATUS_RECTYPE_SMTP;
    ++cProps;
    
    if (msg.getSubject() != NULL) {
        rgprops[cProps].ulPropTag = PR_SUBJECT;
        subj = toWideChar(msg.getSubject());
        rgprops[cProps].Value.lpszW = subj;
        ++cProps;
    }
    if (msg.getFrom() != NULL) {
        rgprops[cProps].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
        from = toWideChar(msg.getFrom());
        rgprops[cProps].Value.lpszW = from;
        ++cProps;
    }
  
    FILETIME ft;

    // Sent time (Date: header). to view internally. this is right
    bt2ft(msg.getDate(), ft);
    rgprops[cProps].ulPropTag = PR_MESSAGE_DELIVERY_TIME; // PR_LAST_MODIFICATION_TIME;
    rgprops[cProps].Value.ft = ft;
    ++cProps;
    
    // Received time (Received: header) to view externally. currently it isn't of use
    //bt2ft(msg.getReceived(), ft);
    //rgprops[cProps].ulPropTag = PR_LAST_MODIFICATION_TIME; // PR_MESSAGE_DELIVERY_TIME;
    //rgprops[cProps].Value.ft = ft;
    //++cProps;

    
    // TODO: MODIFY THE  rgprops TOO!|!!!!!!!!
    //
    rgprops[cProps].ulPropTag = PR_CUSTOM_MESSAGE_ID;
    messId = toWideChar(msg.getMessageId());
    if (messId) {
        rgprops[cProps].Value.lpszW  = messId;
        ++cProps;       
    }

    rgprops[cProps].ulPropTag = PR_CUSTOM_MESSAGE_HEADER;
    h = msg.getHeaders();
    headers = toWideChar(h);
    if (headers) {
        rgprops[cProps].Value.lpszW  = headers;
        ++cProps;       
    }    
  
    // Body  
    bp = msg.getBody();      
    if (bp.getContent()) {
        // Stream the body in...
        hr = pmsg->OpenProperty(PR_BODY, NULL, 0, MAPI_MODIFY | MAPI_CREATE, 
          (LPUNKNOWN*)&pstm);
        EXIT_ON_FAILED(hr);
        wchar_t *content = toWideChar(bp.getContent(), bp.getCharset());
        pstm->Write(content, (wcslen(content)+1)*sizeof(wchar_t), NULL);
        pstm->Release();
        delete [] content;
    }
  
  // Attachment
  AddAttachment(pmsg, m);

  hr = pmsg->SetProps(cProps, rgprops, NULL);
  EXIT_ON_FAILED(hr);

FuncExit:
    if(subj)
        delete [] subj;
    if(from)
        delete [] from;
    if (messId) {
         delete [] messId; messId = NULL;
    }
    if (headers) {
         delete [] headers; headers = NULL;
    }
    if (h) {
        delete [] h; h = NULL;
    }
    LOG.debug("Exiting from SetMessageProps function");
    return hr;
}

void MailServer2Client::setFolderToSync(const wchar_t* t){
    if (server_folderToSync) {
        delete [] server_folderToSync; server_folderToSync = NULL;
    }
    server_folderToSync = wstrdup(t);
}

const wchar_t* MailServer2Client::getFolderToSync() {
    return server_folderToSync;
}

/*
* Return the ID of the inserted message
*/

wchar_t* MailServer2Client::addMessage(MailClientData* m) {

    LOG.debug("Enter into addMessage function");
    HRESULT           hr;
    IMAPISession *    pSession    =  NULL;
    IMAPITable*       pTable      =  NULL;
    SRowSet*          psrs        =  NULL;
    IMsgStore*        pStore      =  NULL;
    int               i           =  0;
    ULONG*   rgTags;
    folderNumber = 0;
    BOOL accountFound             = FALSE;
    if (server_folderToSync == NULL) {
        rgTags = new ULONG[2];
        rgTags[0] = 1; 
        rgTags[1] = PR_CE_IPM_INBOX_ENTRYID; 
        folderNumber = 1;
    }
    else {
        folderNumber = wcslen(server_folderToSync);
        rgTags = new ULONG[folderNumber + 1];         
        rgTags[0] = folderNumber;
        for (int i = 1; i <= folderNumber; i++) 
            rgTags[i] = getFolderId(server_folderToSync[i-1]);    // PR_CE_IPM_INBOX_ENTRYID; 
                                                      // PR_CE_IPM_DRAFTS_ENTRYID;        
    }
    
    ULONG             rgMsgTags[] =  { 1, PR_ENTRYID };
    ULONG             cValues     =  0;
    ULONG             returned    =  0;
    
    SPropValue*       rgprops     =  NULL;
    SPropValue*       rgpropsMsg  =  NULL;
    LPMAPIFOLDER      pfldrInbox  =  NULL;

    IMessage*         pmsg        =  NULL;
    ULONG             cmsgValue   =  0;
    SPropValue*       msgprops    =  NULL;
            
	LPSRowSet	pRows = NULL;    
    SPropValue*       rgprops2    =  NULL;
    ULONG             cValues2    =  0;
    wchar_t* ret                  = NULL;
	       
    // First log on to the store.
    hr = MAPILogonEx (NULL, NULL, NULL, NULL, &pSession);
    EXIT_ON_FAILED (hr);

    // Get the message stores table
    hr = pSession->GetMsgStoresTable (MAPI_UNICODE, &pTable);
    EXIT_ON_FAILED (hr);
   
    hr = pTable->SetColumns((LPSPropTagArray)&Columns, 0);
    EXIT_ON_FAILED(hr);
  
    bool msgset = false; 

    while (!msgset){
        // Get a row
        hr = pTable->QueryRows (1, 0, &psrs);
        EXIT_ON_FAILED (hr);
        
        // Did we hit the end of the table?
        if (psrs->cRows != 1) {
          break;
        }
        //
        // the display name is the name of the account.
        //
        wchar_t* displayName = psrs->aRow[0].lpProps[eePR_DISPLAY_NAME].Value.lpszW;
        
        //
        // The name of the MessageStore that we use for the own Transport
        //
        if (wcscmp(displayName, ACCOUNT_NAME) == 0) {
            accountFound = TRUE;
            // Make sure we got the props we need
            if ((psrs->aRow[0].cValues < 1) || (psrs->aRow[0].lpProps[0].ulPropTag != PR_ENTRYID))
            {
              logError(TEXT("Store missing PR_ENTRYID!"));
              hr = E_FAIL;
              break;
            }

            // Open this message store
            hr = pSession->OpenMsgStore (NULL, 
                           psrs->aRow[0].lpProps[0].Value.bin.cb, 
                          (ENTRYID *) psrs->aRow[0].lpProps[0].Value. 
                           bin.lpb, NULL, 0, &pStore);
            EXIT_ON_FAILED (hr);
      
            // Now get the Inbox folder.
            hr = pStore->GetProps ((SPropTagArray *) rgTags, MAPI_UNICODE, &cValues, &rgprops);
            EXIT_ON_FAILED (hr);
            
            ASSERT (rgprops);
            
            //
            // watch at the folder that we want to sync
            //
            for (i = 0; i < folderNumber; i++) {
                wchar_t* tmp = c2wc(server_folderToSync[i]);

                if (wcscmp(tmp, m->getParent()) != 0) {
                    if (tmp) delete [] tmp; tmp = NULL;
                    continue;  }
            
                // open the message store
                hr = pStore->OpenEntry (rgprops[i].Value.bin.cb, 
                            (LPENTRYID)rgprops[i].Value.bin.lpb, 
                            NULL, 
                            MAPI_MODIFY, 
                            NULL, 
                            (IUnknown **) &pfldrInbox);
                EXIT_ON_FAILED (hr);
            
                if (!pfldrInbox){
                    LOG.error("MailServer2Client::addMessage: pfldrInbox NULL!");
                    goto FuncExit;
                }
                
                if (doesMessageExist(m->getEmailData()->getEmailItem().getMessageId(), pfldrInbox)) {
                    ret = ITEM_ARLEADY_EXISTS;

                } else {
                    hr = pfldrInbox->CreateMessage (NULL, 0, &pmsg);
                    EXIT_ON_FAILED (hr);
          
                    if (!pmsg){
                        LOG.error("MailServer2Client::addMessage: pmsg NULL!");
                        goto FuncExit;
                    }
                                    
                    // Now set the props on the message....
                    hr = SetMessageProps (pmsg, m);
                    EXIT_ON_FAILED(hr);
                    
                    LOG.debug("Message props set.");

                    //
                    // Get the ID of the new message
                    //
                    hr = pmsg->GetProps ((SPropTagArray *) rgMsgTags, MAPI_UNICODE, &cValues2, &rgprops2);
                    EXIT_ON_FAILED(hr);
                    
                    if (rgprops2[eePR_ENTRYID].ulPropTag == PR_ENTRYID) {
                        ret = convertBinaryToWChar(rgprops2[eePR_ENTRYID].Value.bin, server_folderToSync[i]);
                        LOG.debug("Message enry id: %ls", ret);
                    }
                    else {
                        LOG.error("Error getting msg entry id.");
                    }
                    msgset = true;
                    break;  // found.
                }
            }        
        }
                
        // Clean up
        MAPIFreeBuffer (rgprops);    
        MAPIFreeBuffer (rgprops2);    
        MAPIFreeBuffer (rgpropsMsg);  
        FreeProws   (psrs);     
        
        rgprops   =  NULL;  
        rgpropsMsg =  NULL; 
        psrs    =  NULL;    

        RELEASE_OBJ (pmsg);
        RELEASE_OBJ (pfldrInbox);
        RELEASE_OBJ (pStore);
    }
    if (accountFound == FALSE) {
        logError(TEXT("No Funambol account found"));
    }
FuncExit:
  
    MAPIFreeBuffer (rgprops);
    MAPIFreeBuffer (rgpropsMsg);
    FreeProws   (psrs);

    RELEASE_OBJ (pmsg);
    RELEASE_OBJ (pfldrInbox);
    RELEASE_OBJ (pStore);
    RELEASE_OBJ (pTable);
    RELEASE_OBJ (pSession);

    if (rgTags)
        delete rgTags;
    LOG.debug("Exiting from addMessage function");
    return ret;
}


/*
* Return the ID of the inserted message
*/

HRESULT MailServer2Client::updateMessage(MailClientData* m) {
    LOG.debug("Enter into updateMessage function");
    HRESULT           hr;
    IMAPISession *    pSession    =  NULL;
    IMAPITable*       pTable      =  NULL;
    SRowSet*          psrs        =  NULL;
    IMsgStore*        pStore      =  NULL;
    int               i           =  0;
    ULONG*   rgTags;
    folderNumber = 0;
    BOOL accountFound             = FALSE;

    if (server_folderToSync == NULL) {
        rgTags = new ULONG[2];
        rgTags[0] = 1; 
        rgTags[1] = PR_CE_IPM_INBOX_ENTRYID; 
        folderNumber = 1;
    }
    else {
        folderNumber = wcslen(server_folderToSync);
        rgTags = new ULONG[folderNumber + 1];         
        rgTags[0] = folderNumber;
        for (int i = 1; i <= folderNumber; i++) 
            rgTags[i] = getFolderId(server_folderToSync[i-1]);    // PR_CE_IPM_INBOX_ENTRYID; 
                                                      // PR_CE_IPM_DRAFTS_ENTRYID;        
    }
    
    ULONG             rgMsgTags[] =  { 1, PR_ENTRYID };
    ULONG             cValues     =  0;
    ULONG             returned    =  0;
    
    SPropValue*       rgprops     =  NULL;
    SPropValue*       rgpropsMsg  =  NULL;
    LPMAPIFOLDER      pfldrInbox  =  NULL;

    IMessage*         pmsg        =  NULL;
    ULONG             cmsgValue   =  0;
    SPropValue*       msgprops    =  NULL;
            
	LPSRowSet	      pRows        = NULL;    
    SPropValue*       rgprops2     =  NULL;
    ULONG             cValues2     =  0;
    wchar_t*          ret         = NULL;
    wchar_t*          displayName  = NULL;
	       
    // First log on to the store.
    hr = MAPILogonEx (NULL, NULL, NULL, NULL, &pSession);
    EXIT_ON_FAILED (hr);

    // Get the message stores table
    hr = pSession->GetMsgStoresTable (MAPI_UNICODE, &pTable);
    EXIT_ON_FAILED (hr);
   
    hr = pTable->SetColumns((LPSPropTagArray)&Columns, 0);
    EXIT_ON_FAILED(hr);
  
    while (1){
        // Get a row
        hr = pTable->QueryRows (1, 0, &psrs);
        EXIT_ON_FAILED (hr);
        
        // Did we hit the end of the table?
        if (psrs->cRows != 1) {
            break;
        }
        //
        // the display name is the name of the account.
        //
        displayName = psrs->aRow[0].lpProps[eePR_DISPLAY_NAME].Value.lpszW;
        
        //
        // The name of the MessageStore that we use for the own Transport
        //
        if (wcscmp(displayName, ACCOUNT_NAME) == 0) {
            accountFound = TRUE;
            // Make sure we got the props we need
            if ((psrs->aRow[0].cValues < 1) || (psrs->aRow[0].lpProps[0].ulPropTag != PR_ENTRYID))
            {
                MessageBox (NULL, L"Store missing PR_ENTRYID!", L"Doh", MB_OK);
                hr = E_FAIL;
                break;
            }

            // Open this message store
            hr = pSession->OpenMsgStore (NULL, 
                           psrs->aRow[0].lpProps[0].Value.bin.cb, 
                          (ENTRYID *) psrs->aRow[0].lpProps[0].Value. 
                           bin.lpb, NULL, 0, &pStore);
            EXIT_ON_FAILED (hr);
      
            // Now get the Inbox folder.
            hr = pStore->GetProps ((SPropTagArray *) rgTags, MAPI_UNICODE, &cValues, &rgprops);
            EXIT_ON_FAILED (hr);
      
            ASSERT (rgprops);
                   
            //
            // watch at the folder that we want to sync
            //
            for (i = 0; i < folderNumber; i++) {
                wchar_t* t = c2wc(server_folderToSync[i]);

                if (wcscmp(t, m->getParent()) != 0) {
                    if (t) {
                        delete [] t; t = NULL; 
                    }
                    continue; 
                }
            
                // open the message store
                hr = pStore->OpenEntry (rgprops[i].Value.bin.cb, 
                            (LPENTRYID)rgprops[i].Value.bin.lpb, 
                            NULL, 
                            MAPI_MODIFY, 
                            NULL, 
                            (IUnknown **) &pfldrInbox);
                EXIT_ON_FAILED (hr);
            
                ASSERT (pfldrInbox);  
                            
                const char* entryId = m->getEmailData()->getEmailItem().getEntryID();
                // Skip folder letter + /
                if (startWithFolder(entryId))
                    entryId += 2;
                // Decode the id
                char* id = new char[strlen(entryId)];
                int len = b64_decode(id, entryId);
                id[len] = 0;                                                
                
                ENTRYLIST entryList;
                
                entryList.cValues = 0;
			    MAPIAllocateBuffer(sizeof(SBinary), (LPVOID*)&(entryList.lpbin));
                entryList.lpbin[entryList.cValues].cb = len;
		        MAPIAllocateBuffer(len, (LPVOID*)&(entryList.lpbin[entryList.cValues].lpb) );
		        memmove(entryList.lpbin[entryList.cValues].lpb, id, len);
		        entryList.cValues++;
                
                hr = pSession->OpenEntry(entryList.lpbin[0].cb, 
                    (ENTRYID*)entryList.lpbin[0].lpb, 
                    NULL, 
                    MAPI_BEST_ACCESS, 
                    &returned, 
                    (IUnknown **) &pmsg);

                EXIT_ON_FAILED (hr);        
                if(!pmsg){
                    LOG.error("Pmsg is null. ");
                    goto FuncExit;
                }

                if (id)
                    delete [] id;
                
                // Now set the props on the message....
                hr = SetMessageProps (pmsg, m);
                EXIT_ON_FAILED(hr);                       
               
                MAPIFreeBuffer (entryList.lpbin);    
            }        
        }      
        
        // Clean up
        MAPIFreeBuffer (rgprops);    
        MAPIFreeBuffer (rgprops2);    
        MAPIFreeBuffer (rgpropsMsg);  
        FreeProws   (psrs);     

        rgprops   =  NULL;  
        rgpropsMsg =  NULL; 
        psrs    =  NULL;    

        RELEASE_OBJ (pmsg);
        RELEASE_OBJ (pfldrInbox);
        RELEASE_OBJ (pStore);
    }
    if (accountFound == FALSE) {
        logError(TEXT("No Funambol account found"));
    }
FuncExit:
  
    MAPIFreeBuffer (rgprops);
    MAPIFreeBuffer (rgpropsMsg);
    FreeProws   (psrs);

    RELEASE_OBJ (pmsg);
    RELEASE_OBJ (pfldrInbox);
    RELEASE_OBJ (pStore);
    RELEASE_OBJ (pTable);
    RELEASE_OBJ (pSession);

    if (rgTags)
        delete rgTags;
    LOG.debug("Exiting from updateMessage function");
    return hr;
}



/*
* Return the ID of the inserted message
*/

HRESULT MailServer2Client::deleteMessage(MailClientData* m) {
    LOG.debug("Enter into deleteMessage function");
    HRESULT           hr;
    IMAPISession *    pSession    =  NULL;
    IMAPITable*       pTable      =  NULL;
    SRowSet*          psrs        =  NULL;
    IMsgStore*        pStore      =  NULL;
    int               i           =  0;
    ULONG*   rgTags;
    folderNumber = 0;
    BOOL accountFound             = FALSE;

    if (server_folderToSync == NULL) {
        rgTags = new ULONG[2];
        rgTags[0] = 1; 
        rgTags[1] = PR_CE_IPM_INBOX_ENTRYID; 
        folderNumber = 1;
    }
    else {
        folderNumber = wcslen(server_folderToSync);
        rgTags = new ULONG[folderNumber + 1];         
        rgTags[0] = folderNumber;
        for (int i = 1; i <= folderNumber; i++) 
            rgTags[i] = getFolderId(server_folderToSync[i-1]);    // PR_CE_IPM_INBOX_ENTRYID; 
                                                      // PR_CE_IPM_DRAFTS_ENTRYID;        
    }
    
    ULONG             rgMsgTags[] =  { 1, PR_ENTRYID };
    ULONG             cValues     =  0;
    ULONG             returned    =  0;
    
    SPropValue*       rgprops     =  NULL;
    SPropValue*       rgpropsMsg  =  NULL;
    LPMAPIFOLDER      pfldrInbox  =  NULL;

    IMessage*         pmsg        =  NULL;
    ULONG             cmsgValue   =  0;
    SPropValue*       msgprops    =  NULL;
            
	LPSRowSet	      pRows        = NULL;    
    SPropValue*       rgprops2     =  NULL;
    ULONG             cValues2     =  0;
    wchar_t*          ret         = NULL;
    wchar_t*          displayName  = NULL;
	       
    // First log on to the store.
    hr = MAPILogonEx (NULL, NULL, NULL, NULL, &pSession);
    EXIT_ON_FAILED (hr);

    // Get the message stores table
    hr = pSession->GetMsgStoresTable (MAPI_UNICODE, &pTable);
    EXIT_ON_FAILED (hr);
   
    hr = pTable->SetColumns((LPSPropTagArray)&Columns, 0);
    EXIT_ON_FAILED(hr);
  
    while (1){
        // Get a row
        hr = pTable->QueryRows (1, 0, &psrs);
        EXIT_ON_FAILED (hr);
        
        // Did we hit the end of the table?
        if (psrs->cRows != 1) {
            break;
        }
        //
        // the display name is the name of the account.
        //
        displayName = psrs->aRow[0].lpProps[eePR_DISPLAY_NAME].Value.lpszW;
    
        //
        // The name of the MessageStore that we use for the own Transport
        //
        if (wcscmp(displayName, ACCOUNT_NAME) == 0) {
            accountFound = TRUE;
            // Make sure we got the props we need
            if ((psrs->aRow[0].cValues < 1) || (psrs->aRow[0].lpProps[0].ulPropTag != PR_ENTRYID))
            {
              logError(TEXT("Store missing PR_ENTRYID!"));
              hr = E_FAIL;
              break;
            }

            // Open this message store
            hr = pSession->OpenMsgStore (NULL, 
                           psrs->aRow[0].lpProps[0].Value.bin.cb, 
                          (ENTRYID *) psrs->aRow[0].lpProps[0].Value. 
                           bin.lpb, NULL, 0, &pStore);
            EXIT_ON_FAILED (hr);
      
            // Now get the Inbox folder.
            hr = pStore->GetProps ((SPropTagArray *) rgTags, MAPI_UNICODE, &cValues, &rgprops);
            EXIT_ON_FAILED (hr);
      
            ASSERT (rgprops);
                   
            //
            // watch at the folder that we want to sync
            //
            for (i = 0; i < folderNumber; i++) {
                wchar_t* t = c2wc(server_folderToSync[i]);
              
                if (wcscmp(t, m->getParent()) != 0) {
                    if (t) {
                        delete [] t; t = NULL; 
                    }
                    continue; 
                }
            
                // open the message store
                hr = pStore->OpenEntry (rgprops[i].Value.bin.cb, 
                            (LPENTRYID)rgprops[i].Value.bin.lpb, 
                            NULL, 
                            MAPI_MODIFY, 
                            NULL, 
                            (IUnknown **) &pfldrInbox);
                EXIT_ON_FAILED (hr);
            
                ASSERT (pfldrInbox);  
                            
                const char* entryId = m->getEmailData()->getEmailItem().getEntryID();
                // Skip folder letter + /
                if (startWithFolder(entryId))
                    entryId += 2;
                // Decode the id
                char* id = new char[strlen(entryId)];
                int len = b64_decode(id, entryId);
                id[len] = 0;                                                

                ENTRYLIST entryList;
                
                entryList.cValues = 0;
			    MAPIAllocateBuffer(sizeof(SBinary), (LPVOID*)&(entryList.lpbin));
                entryList.lpbin[entryList.cValues].cb = len;
		        MAPIAllocateBuffer(len, (LPVOID*)&(entryList.lpbin[entryList.cValues].lpb) );
		        memmove(entryList.lpbin[entryList.cValues].lpb, id, len);
		        entryList.cValues++;
                
                if (id) 
                    delete id;

                hr = pfldrInbox->DeleteMessages (&entryList, 0, NULL, 0);
                EXIT_ON_FAILED (hr);
               
                MAPIFreeBuffer (entryList.lpbin);    
            }        
        }        
        // Clean up
        MAPIFreeBuffer (rgprops);    
        MAPIFreeBuffer (rgprops2);    
        MAPIFreeBuffer (rgpropsMsg);  
        FreeProws   (psrs);     
        
        rgprops   =  NULL;  
        rgpropsMsg =  NULL; 
        psrs    =  NULL;    

        RELEASE_OBJ (pmsg);
        RELEASE_OBJ (pfldrInbox);
        RELEASE_OBJ (pStore);
    }
    if (accountFound == FALSE) {
        logError(TEXT("No Funambol account found"));
        goto FuncExit;
    }
FuncExit:
  
    MAPIFreeBuffer (rgprops);
    MAPIFreeBuffer (rgpropsMsg);
    FreeProws   (psrs);

    RELEASE_OBJ (pmsg);
    RELEASE_OBJ (pfldrInbox);
    RELEASE_OBJ (pStore);
    RELEASE_OBJ (pTable);
    RELEASE_OBJ (pSession);

    if (rgTags)
        delete rgTags;
    LOG.debug("Exiting from deleteMessage function");
    return hr;
}

bool MailServer2Client::doesMessageExist (IN const char* msgid, LPMAPIFOLDER pfldrInbox) {
	
    if (msgid == NULL) 
        return false;
    
    wchar_t* msgID = toWideChar(msgid);
	HRESULT hr = S_OK;
	bool bFound = false;
	SBinary	sbEntry = {0};
	LPMAPITABLE pContentsTable = NULL;
	LPSRowSet pRows = NULL;
	SRow aRow = {0};
	
	enum {
		ePR_ENTRYID,			// 0		
		ePR_CUSTOM_MESSAGE_ID,			// 2
		NUM_COLS				// number of columns (put it to the end)
	};

	// These tags represent the message information we would like to pick up
	static SizedSPropTagArray(NUM_COLS, Columns) =
	{
		NUM_COLS,
		PR_ENTRYID,
		PR_CUSTOM_MESSAGE_ID
	};

	hr = pfldrInbox->GetContentsTable(0, &pContentsTable);
	EXIT_ON_FAILED(hr);

	hr = pContentsTable->SetColumns((LPSPropTagArray)&Columns, 0);
	EXIT_ON_FAILED(hr);

	// iterate message one by one (each time only read 1)
	while(!FAILED(hr = pContentsTable->QueryRows(1, 0, &pRows))) {
		// this is the end of the content table
		if(pRows->cRows == 0) break;

		// check this Row
		aRow = pRows->aRow[0];

		// Correctly returned?
		if ((&aRow.lpProps[ePR_ENTRYID])->ulPropTag == PR_ENTRYID			
			&& (&aRow.lpProps[ePR_CUSTOM_MESSAGE_ID])->ulPropTag == PR_CUSTOM_MESSAGE_ID) {

			// Is CUSTOM_MESSAGE_ID found?
            wchar_t* mm = aRow.lpProps[ePR_CUSTOM_MESSAGE_ID].Value.lpszW;
            if (wcscmp(mm, msgID) == 0) {
				bFound = true;
				break;
			}
		} // if

		// Free pRows for each row
		FreeProws(pRows);
		pRows = NULL;
	} // while	

FuncExit:
	if(pRows) {
		// Free each element in the pRows buffer
		FreeProws(pRows);
		pRows = NULL;
	}
	RELEASE_OBJ(pContentsTable);
    
    if (msgID)  {
        delete [] msgID; msgID = NULL;
    }

	return bFound;
}