/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "pim/mail/MailUtils.h"
#include "notify/checknet.h"

wchar_t* c2wc (wchar_t t) {    
    wchar_t* ret = new wchar_t[2];
    wsprintf(ret, TEXT("%c"), t);
    return ret;
}

wchar_t* convertBinaryToWChar(SBinary sbEntry, wchar_t folder) {
       
    int len = sbEntry.cb;
    char* b64 = new char[((len/3+1)<<2) + 1];
    len = b64_encode(b64, sbEntry.lpb, len);
    b64[len] = 0;
    wchar_t* t = toWideChar(b64);  
    wchar_t* ret = new wchar_t[wcslen(t) + 1 + 2];
    wsprintf(ret, TEXT("%c/%s"), folder, t);
        
    // AAAAAG8TAAECAAAAAQAAA==
    delete [] b64; b64 = NULL;
    delete [] t; t = NULL;
    return ret;
}

char* convertBinaryToChar(SBinary sbEntry, wchar_t folder) {
       
    int len = sbEntry.cb;
    char* b64 = new char[((len/3+1)<<2) + 1];
    len = b64_encode(b64, sbEntry.lpb, len);
    b64[len] = 0;
    char* ret = new char[strlen(b64) + 1 + 2];
    sprintf(ret, "%lc/%s", folder, b64);
        
    // AAAAAG8TAAECAAAAAQAAA==
    delete [] b64; b64 = NULL;
    return ret;
}


ULONG getFolderId(wchar_t t) {
    
    ULONG ret = -1;

    if (t == INBOX_FOLDER) 
        ret = PR_CE_IPM_INBOX_ENTRYID;
    else if (t == OUTBOX_FOLDER)
         ret = PR_IPM_OUTBOX_ENTRYID;
    else if (t == TRASH_FOLDER)
         ret = PR_IPM_WASTEBASKET_ENTRYID;
    else if (t == DRAFT_FOLDER)
         ret = PR_CE_IPM_DRAFTS_ENTRYID;
    else if (t == SENT_FOLDER)
         ret = PR_IPM_SENTMAIL_ENTRYID;   
    return ret;
}


void getFolderToSync(MailSyncSourceConfig &sc, OUT wchar_t* tt) {
    int i = 0;

    if (sc.getInbox()) {
        tt[i] = INBOX_FOLDER;
        i++;
    }
    if (sc.getOutbox()) {
        tt[i] = OUTBOX_FOLDER;
        i++;
    }
    if (sc.getDraft()) {
        tt[i] = DRAFT_FOLDER;
        i++;
    }   
    if (sc.getSent()) {
        tt[i] = SENT_FOLDER;
        i++;
    }
    if (sc.getTrash()) {
        tt[i] = TRASH_FOLDER;
        i++;
    }
    tt[i] = 0;

}

/*
* In the extras ArrayList there is MailExtraData in which there is the extra info
* as the message-id. Note the function return only the pointer to the class.
* It MUST NOT be released by the caller!!
* 
*/
const MailExtraData* getMailExtraData(ArrayList& extras, const char* entryID) {

    if (extras.isEmpty())
        return NULL;
    
    MailExtraData* med = NULL;
    int i = 0;
    
    for (i = 0; i < extras.size(); i++) {
        
        if (strstr( ((MailExtraData*)extras[i])->getId(), entryID) != NULL) {
            med = (MailExtraData*)extras[i];
            break;
        }        
    }
    
    return med;

}

char* createMessageID(const char* entryID) {
    const char* address ;

    if(checkNetwork())
        address = getLocalAddress();
    else
        address = "127.0.0.1";
    
    char* messageID = new char[strlen(entryID)+strlen(address)+2];

    sprintf(messageID, "%s@%s", entryID, address);
    
    return messageID;
}

